SystemOrganization addCategory: #Loom!

Object subclass: #LoomProfiler
	instanceVariableNames: 'profiledThreads beginingProfilingTime profilingTime profilingProcess numberOfSamplings'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Loom'!

!LoomProfiler methodsFor: 'accessing computed' stamp: 'AlexandreBergel 9/5/2012 11:58'!
consumptionTimeOf: aSmalltalkProcess
	^ ((self getThreadFrom: aSmalltalkProcess) tick / numberOfSamplings) asFloat! !

!LoomProfiler methodsFor: 'private' stamp: 'AlexandreBergel 9/5/2012 11:38'!
createProfilingThread
	profilingProcess := [ [ self isProfiling ] whileTrue: [ 
							self sampleExecution.
							(Delay forMilliseconds: 1) wait.
						 ] ] fork.
! !

!LoomProfiler methodsFor: 'private' stamp: 'AlexandreBergel 9/5/2012 10:57'!
getLoomThreadsFrom: processes
	"Return a collection of LoomThreads from a list of Smalltalk processes"

	| id |
	id := 0.
	^ processes collect: [ :p |
		id := id + 1.
		LoomThread new
			process: p;
			id: id;
			yourself ]! !

!LoomProfiler methodsFor: 'public' stamp: 'AlexandreBergel 9/5/2012 11:11'!
getThreadFrom: aSmalltalkProcess
	"Return the loom thread associated to the smalltalk process"
	
	^ profiledThreads detect: [ :t | t process == aSmalltalkProcess ]! !

!LoomProfiler methodsFor: 'public' stamp: 'AlexandreBergel 9/5/2012 11:49'!
getThreadFrom: aSmalltalkProcess ifAbsent: aBlock
	"Return the loom thread associated to the smalltalk process"
	
	^ profiledThreads detect: [ :t | t process == aSmalltalkProcess ] ifNone: aBlock! !

!LoomProfiler methodsFor: 'initialize-release' stamp: 'AlexandreBergel 9/5/2012 11:57'!
initialize
	super initialize.
	profilingTime := 0.
	beginingProfilingTime := 0.
	numberOfSamplings := 0! !

!LoomProfiler methodsFor: 'testing' stamp: 'AlexandreBergel 9/5/2012 11:41'!
isProfiling
	"Are we profiling?"
	
	^ beginingProfilingTime ~= 0! !

!LoomProfiler methodsFor: 'public' stamp: 'AlexandreBergel 9/5/2012 10:50'!
profile: processCollection
	profiledThreads := self getLoomThreadsFrom: processCollection.
	self startProfiling! !

!LoomProfiler methodsFor: 'accessing' stamp: 'AlexandreBergel 9/5/2012 10:36'!
profilingTime
	^ profilingTime! !

!LoomProfiler methodsFor: 'private' stamp: 'AlexandreBergel 9/5/2012 11:57'!
sampleExecution 
	"Simply increases the tick of the loom thread next to be executed"

	| loomThread |
	loomThread := self getThreadFrom: Processor nextReadyProcess ifAbsent: [ ^ self ].
	loomThread increaseTick.
	numberOfSamplings := numberOfSamplings + 1! !

!LoomProfiler methodsFor: 'public' stamp: 'AlexandreBergel 9/5/2012 11:25'!
startProfiling
	self startTimer.
	self createProfilingThread! !

!LoomProfiler methodsFor: 'private' stamp: 'AlexandreBergel 9/5/2012 11:25'!
startTimer
	beginingProfilingTime := Time now asMilliSeconds! !

!LoomProfiler methodsFor: 'public' stamp: 'AlexandreBergel 9/5/2012 11:39'!
stopProfiling
	profilingTime := Time now asMilliSeconds - beginingProfilingTime.
	beginingProfilingTime := 0! !

Object subclass: #LoomThread
	instanceVariableNames: 'process id tick'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Loom'!

!LoomThread methodsFor: 'accessing' stamp: 'AlexandreBergel 9/5/2012 10:58'!
id
	"Return the id of the thread. If 0, then it means that none has been set"

	^ id! !

!LoomThread methodsFor: 'accessing' stamp: 'AlexandreBergel 9/5/2012 10:58'!
id: aNumber
	"Set the id number"
	
	id := aNumber! !

!LoomThread methodsFor: 'public' stamp: 'AlexandreBergel 9/5/2012 11:50'!
increaseTick
	tick := tick + 1! !

!LoomThread methodsFor: 'initialize' stamp: 'AlexandreBergel 9/5/2012 11:50'!
initialize
	super initialize.
	id := 0.
	tick := 0! !

!LoomThread methodsFor: 'accessing' stamp: 'AlexandreBergel 9/5/2012 10:58'!
process
	^ process! !

!LoomThread methodsFor: 'accessing' stamp: 'AlexandreBergel 9/5/2012 10:58'!
process: anObject
	process := anObject! !

!LoomThread methodsFor: 'accessing' stamp: 'AlexandreBergel 9/5/2012 11:51'!
tick
	^ tick! !

TestCase subclass: #LoomTest
	instanceVariableNames: 'thread1 thread2 profiler thread3'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Loom'!

!LoomTest methodsFor: 'util' stamp: 'AlexandreBergel 9/5/2012 14:23'!
isScenarioOver
	(Array with: thread1 with: thread2 with: thread3) 
		do: [ :t | (t notNil and: [ t isTerminated not ]) ifTrue: [ ^ false ] ].
	^ true
! !

!LoomTest methodsFor: 'scenarios' stamp: 'AlexandreBergel 9/5/2012 09:54'!
scenario1

	thread1 := [ 5 timesRepeat: [ self wait ] ] fork.
	thread2 := [ 5 timesRepeat: [ self wait2 ] ] fork.! !

!LoomTest methodsFor: 'scenarios' stamp: 'AlexandreBergel 9/5/2012 14:27'!
scenario2

	thread1 := [ 200 timesRepeat: [ self wait ] ] fork.
	thread2 := [ 100 timesRepeat: [ self wait2 ] ] fork.
	thread3 := [ 50 timesRepeat: [ self wait3 ] ] fork.! !

!LoomTest methodsFor: 'running' stamp: 'AlexandreBergel 9/5/2012 10:51'!
setUp
	super setUp.
	profiler := LoomProfiler new.! !

!LoomTest methodsFor: 'running' stamp: 'AlexandreBergel 9/5/2012 10:54'!
tearDown
	profiler stopProfiling! !

!LoomTest methodsFor: 'tests' stamp: 'AlexandreBergel 9/5/2012 11:58'!
testBasic

	self scenario1.
	profiler profile: (Array with: thread1 with: thread2).
	self waitForScenarioEnd.
	profiler stopProfiling.
	
	self assert: profiler profilingTime > 10.
	
	"Thread1 consume twice more the CPU than Thead2. Actually, thread2 is sleeping twice much."
	self assert: (profiler consumptionTimeOf: thread1) > 0.4.
	self assert: (profiler consumptionTimeOf: thread2) < 0.6.
! !

!LoomTest methodsFor: 'tests' stamp: 'AlexandreBergel 9/5/2012 14:32'!
testBasic2

	self scenario2.
	profiler profile: (Array with: thread1 with: thread2 with: thread3).
	self waitForScenarioEnd.
	profiler stopProfiling.
	
	"Thread1 consume twice more the CPU than Thead2. Actually, thread2 is sleeping twice much."
	self assert: (profiler consumptionTimeOf: thread1) >= 0.6.
	self assert: (profiler consumptionTimeOf: thread1) < 1.0.

	self assert: (profiler consumptionTimeOf: thread2) >= 0.2.
	self assert: (profiler consumptionTimeOf: thread2) < 0.6.
	
	self assert: (profiler consumptionTimeOf: thread3) > 0.0.
	self assert: (profiler consumptionTimeOf: thread3) < 0.3.
! !

!LoomTest methodsFor: 'tests' stamp: 'AlexandreBergel 9/5/2012 11:13'!
testGetLoomThreadFrom
	| p ps |
	p := [ ] fork.

	profiler profile: (Array with: p).
	self assert: (profiler getThreadFrom: p) process == p.
	self assert: (profiler getThreadFrom: p) id = 1! !

!LoomTest methodsFor: 'tests' stamp: 'AlexandreBergel 9/5/2012 11:11'!
testGetLoomThreads
	| p ps |
	p := [ ] fork.

	ps := profiler getLoomThreadsFrom: (Array with: p).

	self assert: ps size = 1.
	self assert: ps first process == p.
	self assert: ps first id = 1.
! !

!LoomTest methodsFor: 'tests' stamp: 'AlexandreBergel 9/5/2012 11:40'!
testIsProfiling

	self deny: profiler isProfiling.
	profiler startProfiling.
	self assert: profiler isProfiling.
	
	profiler stopProfiling.
	self deny: profiler isProfiling.
	
! !

!LoomTest methodsFor: 'util' stamp: 'AlexandreBergel 9/5/2012 09:55'!
wait
	(Delay forMilliseconds: 2) wait! !

!LoomTest methodsFor: 'util' stamp: 'AlexandreBergel 9/5/2012 09:55'!
wait2
	(Delay forMilliseconds: 4) wait! !

!LoomTest methodsFor: 'util' stamp: 'AlexandreBergel 9/5/2012 14:25'!
wait3
	(Delay forMilliseconds: 8) wait! !

!LoomTest methodsFor: 'util' stamp: 'AlexandreBergel 9/5/2012 10:12'!
waitForScenarioEnd
	[ self isScenarioOver not ] whileTrue: [ self wait ]! !
